/***************************************************************************
			      z42_cmyk.c
                         -------------------
    version		 : 0.4.0
    date	         : Wed Sep 10 2003
    copyright            : (C) 2001,2003 by Alwin Stolk
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

/***************************************************************************
 Make cmyk file:
    gs -dSAFER -r300x600 -sPAPERSIZE=a4 -q -dNOPAUSE -dBATCH \
    -sDEVICE=bitcmyk -sOutputFile=output.cmyk nicepicture.ps

 Send to printer:
    z42_cmyk size=a4 hdpi=300 mode=cmyk <output.cmyk >/dev/printer
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#define BUF_SZ		(2*150000L) // (packets+GAP)*RAW_PACKET_SZ ; @ 1200 dpi => full width @@@ TODO: malloc

#define HEADER_SZ	34
#define RAW_PACKET_SZ	28
#define PACKET_HDR_SZ	4

#define NOZZLES_BLACK	192
#define NOZZLES_COLOUR  64

#define BLACK		1
#define YELLOW		2
#define MAGENTA		4
#define CYAN		8

#define BORDER_BOTTOM	41	// pixels @ 72 dpi
#define BORDER_RIGHT	3	// lexmark driver: TOP:1mm BOTTOM:14.5mm LEFT/RIGHT:4mm

#define TRUE            1
#define FALSE           !TRUE


enum printmodes {
	MODE_BLACK, MODE_CMY, MODE_CMYK
};


typedef unsigned char bool;

typedef struct {
	char *name;
	unsigned width;
	unsigned height;
} papersize;


char *modenames[] = { "k", "cmy", "cmyk", NULL };

// @ 0x18: alignment parameter 'b' ; not used here?
unsigned char leadin[] = {
	0x1B,0x2A,0x81,0x00,0x1C,0x50,0x41,0x00,0x01,0x00,0x58,0x02,0x04,0x00,0xC0,0x12,
	0xC8,0x19,0x02,0x00,0x50,0x00,0x14,0x00,0x07,0x00,0x08,0x00,0x1B,0x2A,0x07,0x73,
	0x30,0x1B,0x2A,0x6D,0x00,0x14,0x01,0xC0,0x02,0x00,0x01,0xBE,0x1B,0x2A,0x07,0x63,
};
unsigned char leadout[] = {
	0x1B,0x2A,0x07,0x65,0x1B,0x2A,0x82,0x00,0x00,0x00,0x00,0xAC
};

// units = pixels @ 72 dpi
// width  3"..8.5"
// height 5"..17"
papersize ghostscript[] = {
	{ "a4", 595, 842 },
	{ "a5", 421, 595 },
	{ "a6", 297, 421 },
	{ "legal", 612, 1008 },
	{ "letter", 612, 792 },
	{ "isob5", 501, 709 },
	{ "isob6", 354, 501 },
	{ "jisb5", 516, 729 },
	{ "jisb6", 363, 516 },
	{ "c5", 459, 649 },
	{ "c6", 323, 459 },
	{ "flsa", 612, 936 },
	{ "flse", 612, 936 },
	{ "halfletter", 396, 612 },
	{ NULL, 0, 0 }
};

FILE *out=NULL;
FILE *in=NULL;
unsigned gs_width=595, gs_height=842; // default papersize is a4
unsigned hdpi=600; // default horizontal dpi
unsigned gap; // distance between the two rows of nozzles
unsigned char direction=1;
bool bidirectional=TRUE;
short mode=MODE_BLACK;

/***************************************************************************/

void convert_dpi(unsigned *val, unsigned from, unsigned to)
{
	float x;

	x = (float)*val * to / from;
	*val = (unsigned) x;
	x -= *val;
	if (x >= 0.5f) (*val)++;
}


void put_block(char *p, long len)
{
	fwrite(p, 1, len, out);
}


bool buffer_used(char *p)
{
	long x;

	for (x=0; x<BUF_SZ; x++) if (*p++) return TRUE;
	return FALSE;
}


// use for bi-directional printing
void reverse_buffer(unsigned char *p, unsigned packets)
{
	unsigned i;
	unsigned char tmp[RAW_PACKET_SZ];

	for (i=0; i<(packets/2); i++) {
		memcpy(tmp, p+i*RAW_PACKET_SZ, RAW_PACKET_SZ);
		memcpy(p+i*RAW_PACKET_SZ, p+(packets-1-i)*RAW_PACKET_SZ, RAW_PACKET_SZ);
		memcpy(p+(packets-1-i)*RAW_PACKET_SZ, tmp, RAW_PACKET_SZ);
	}
}


long compress_buffer(unsigned char *p, unsigned packets)
{
	long x, y, i=0, imask;
	unsigned char dat, *pack;

	// only 192 nozzles supported.
	// In 208 nozzle mode bits 0&1 of byte 0 are used as mask also
	for(x=0; x<packets; x++) {
		imask = i+1;
		p[i++] = 0x80;
		p[i++] = 0;
		p[i++] = 0;
		p[i++] = 0;

		pack = p + x*RAW_PACKET_SZ + PACKET_HDR_SZ;
		for(y=0; y<24; y++) {
			dat = *pack++;
			if (dat) {
				p[i++] = dat;
			} else {
				p[imask + y/8] |= (0x80 >> (y&7));
			}
		}

		if (i&1) p[i++] = 0; // packets always have even length
	}

	return i;
}


char *crop_buffer(unsigned char *p, unsigned *packets, unsigned *offset)
{
        unsigned y, x=0;
	unsigned char *pack;

        while (*packets) {
		pack = (p + (((*packets)-1)*RAW_PACKET_SZ) + PACKET_HDR_SZ); 
                for (y=0; y<24; y++) {
			if (*pack++) goto cropleft;
                }
                (*packets)--;
        }

cropleft:
        for (x=0; x<(*packets); x++) {
		pack = (p + x*RAW_PACKET_SZ + PACKET_HDR_SZ); 
                for (y=0; y<24; y++) {
			if (*pack++) goto cropdone;
                }
		switch (hdpi) {
			case 300:  (*offset) += 8; break;
			case 600:  (*offset) += 4; break;
			case 1200: (*offset) += 2; break;
			case 2400: (*offset) ++; break;
		}
        }

cropdone:
        (*packets) -= x;
        return (p + x*RAW_PACKET_SZ);
}


void move_paper(int dpi1200)
{
	unsigned char move[] = { 0x1B,0x2A,0x03,0x01,0x20 };

	move[3] = (char)(dpi1200>>8);
	move[4] = (char) dpi1200;
	put_block(move, 5);
}
	

void make_header(unsigned char *header, unsigned long blk_length, unsigned packets, 
	unsigned hstart, unsigned yoff, bool colour)
{
	unsigned hend;

	header[0x00] = 0x1B; // magic 0
	header[0x01] = 0x2A; // magic 1
	header[0x02] = 0x24; // print command

	header[0x03] = (char)(blk_length >> 24);
	header[0x04] = (char)(blk_length >> 16);
	header[0x05] = (char)(blk_length >> 8);
	header[0x06] = (char)(blk_length);

	switch (hdpi) {
		case 300: header[7]=1; break;
		case 600: header[7]=3; break;
		case 1200: header[7]=4; break;
		case 2400: header[7]=5; break;
	}

	header[0x08] = direction; // direction: 1=>left2right 2=>right2left
	if (bidirectional) {
		if (direction==1) direction=2;
		else direction=1;	
	}

	if (colour) {
		header[0x09] = 0x00;
	} else {
		header[0x09] = 0x01;
	}

	header[0x0A] = 0x18; // nozzles: 0x18=>192 0x1A=>208
	header[0x0B] = 0;
	header[0x0C] = 1;    // ?

	header[0x0D] = (char)(packets >> 8);// (packets-1)*4 = hend-hstart
	header[0x0E] = (char)(packets);

	header[0x0F] = (char)(hstart >> 8); // horizontal start BE
	header[0x10] = (char)(hstart);

	hend = hstart + (packets-1)*(2400/hdpi);
	header[0x11] = (char)(hend >> 8); // horizontal end BE
	header[0x12] = (char)(hend);

        header[0x13] = (char)(yoff >> 8); // y offset BE (1200 dpi)
        header[0x14] = (char)(yoff);

	header[0x15] = 0;
	header[0x16] = 0;
	header[0x17] = 0;
	header[0x18] = 0;
	header[0x19] = 0;
	header[0x1A] = 0;
	header[0x1B] = 0;
	header[0x1C] = 0;
	header[0x1D] = 0;
	header[0x1E] = 0;
	header[0x1F] = 0;
	header[0x20] = 0;
	header[0x21] = 0;
}


void print(void)
{
	unsigned char pixel, header[HEADER_SZ];
        unsigned char cbuf0[BUF_SZ], cbuf1[BUF_SZ], cbuf2[BUF_SZ], cbuf3[BUF_SZ];
	unsigned char bbuf0[BUF_SZ], bbuf1[BUF_SZ];
        unsigned x, packets;
	long i, y, y2, length, x_buffer_off, gap_off;
	long offc, offm, offy, offk;
	unsigned char maskc, maskm, masky, maskk;
	unsigned char *bufc, *bufm, *bufy, *bufk, *thedata;
        unsigned width, height, left;
        bool flush;
	unsigned char shift[] = { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 };
	unsigned char lbuf[7500]; // @@@ TODO:malloc

        // set up width and height
	width = gs_width - BORDER_RIGHT;
	height = gs_height - BORDER_BOTTOM;
	convert_dpi(&width, 72, hdpi);
	convert_dpi(&height, 72, 600);
	convert_dpi(&gs_width, 72, hdpi);
	convert_dpi(&gs_height, 72, 600);
	if (width&1) width++;		// ghostscript has even lines. this is good.
	if (gs_width&1) gs_width++;

	// set up resolution specific parameters
	switch (hdpi) {
		case 300: gap=8; break;
		case 600: gap=16; break;
		case 1200: gap=32; break;
		case 2400: gap=64; break;
	}
	gap_off = (long)gap*RAW_PACKET_SZ;

	// print
	while(1) {
                y=0;

                fgetc(in);
                if( feof(in) ) break;
                fseek(in, -1, SEEK_CUR);

		put_block(leadin, 0x30);
		move_paper(0xA0);

		memset(cbuf0, 0, BUF_SZ);
		memset(cbuf1, 0, BUF_SZ);
		memset(cbuf2, 0, BUF_SZ);
		memset(cbuf3, 0, BUF_SZ);
		memset(bbuf0, 0, BUF_SZ);
		memset(bbuf1, 0, BUF_SZ);
		while(y<height) {
			fread(lbuf, 1, gs_width/2, in);
	
			// C
			y2 = y%NOZZLES_COLOUR;
			if (y2 < 12) {
				y2 += 52;
				bufc = cbuf2;
			} else {
				y2 -= 12;
				bufc = cbuf3;
			}
			offc = y2/8;
			if (!(y2&1)) offc += gap_off;
			maskc = shift[y2&7];

			// M
			y2 = y%NOZZLES_COLOUR;
			if (y2 < 38) {
				y2 += 26;
				bufm = cbuf1;
			} else {
				y2 -= 38;
				bufm = cbuf2;
			}
			offm = y2/8;
			if (!(y2&1)) offm += gap_off;
			maskm = shift[y2&7];

			// Y
			bufy = cbuf0;
			offy = (y%NOZZLES_COLOUR)/8;
			if (!(y&1)) offy += gap_off;
			masky = shift[y&7];
		
			// K
			y2 = y%NOZZLES_BLACK;
			if (y2 < 160) {
				y2 += 32;
				bufk = bbuf0;
			} else {
				y2 -= 160;
				bufk = bbuf1;
			}
			offk = y2/8;
			if (!(y2&1)) offk += gap_off;
			maskk = shift[y2&7];

			// pixel line
			for (x=0; x<width; x++) {
				pixel = lbuf[x>>1];
				if (x&1) {
					pixel &= 15;
				} else {
					pixel >>= 4;
				}

				x_buffer_off = (long)x*RAW_PACKET_SZ + PACKET_HDR_SZ;

				switch (mode) {
					case MODE_BLACK:
						pixel &= BLACK;
						break;

					case MODE_CMY:
						// fake black
						if (pixel&BLACK) pixel = CYAN|MAGENTA|YELLOW;
						break;
				}

				if (pixel & BLACK)   bufk[x_buffer_off + offk] |= maskk;
				if (pixel & CYAN)    bufc[x_buffer_off + offc] |= maskc;
				if (pixel & MAGENTA) bufm[x_buffer_off + offm + 8] |= maskm;
				if (pixel & YELLOW)  bufy[x_buffer_off + offy + 16] |= masky;
			}

			y++;
			if (y==height) {
				if (mode==MODE_BLACK) {
					while (y%NOZZLES_BLACK) y++;
				} else {
					while (y%NOZZLES_COLOUR) y++;
				}
			}

			if ( !(y%NOZZLES_COLOUR) ) {
                                do {
					if (mode==MODE_CMYK || mode==MODE_CMY) {
	                                        if ( buffer_used(cbuf0) ) {
							left=0;
							packets = gap + width;
							thedata = crop_buffer(cbuf0, &packets, &left);
							if (direction==2) reverse_buffer(thedata, packets);
                		                        length = HEADER_SZ + compress_buffer(thedata, packets);
							make_header(header, length, packets, 0x10+left, 0, TRUE);
                                		        put_block(header, HEADER_SZ);
	                                	        put_block(thedata, length-HEADER_SZ);
						}
        	                                memcpy(cbuf0, cbuf1, BUF_SZ);
                	                        memcpy(cbuf1, cbuf2, BUF_SZ);
                        	                memcpy(cbuf2, cbuf3, BUF_SZ);
                                	        memset(cbuf3, 0, BUF_SZ);
					}

					if ( (mode==MODE_BLACK || mode == MODE_CMYK) && !(y%NOZZLES_BLACK) ) { 
	                                        if ( buffer_used(bbuf0) ) {
							left=0;
							packets = gap + width;
							thedata = crop_buffer(bbuf0, &packets, &left);
							if (direction==2) reverse_buffer(thedata, packets);
	       		                                length = HEADER_SZ + compress_buffer(thedata, packets);
							make_header(header, length, packets, 0x10+left, 0, FALSE);
	                        	                put_block(header, HEADER_SZ);
	                                	        put_block(thedata, length-HEADER_SZ);
						}
						memcpy(bbuf0, bbuf1, BUF_SZ);
						memset(bbuf1, 0, BUF_SZ);
 					}
					
					flush = buffer_used(cbuf0) | buffer_used(cbuf1) | buffer_used(cbuf2);
					flush |= buffer_used(bbuf0);

					if (mode==MODE_BLACK) {
						if ((y<height || flush) && (!(y%NOZZLES_BLACK))) {
							move_paper(NOZZLES_BLACK*2);
							if (y>=height) y += NOZZLES_BLACK;
						}
					} else {
						if (y<height || flush) {
							move_paper(NOZZLES_COLOUR*2);
							if (y>=height) y += NOZZLES_COLOUR;
						}
					}

                               } while (y>=height && flush);
                        }
		}

                if (height<gs_height) {
                        for (i=0; i<(gs_height-height); i++) fread(lbuf, 1, gs_width/2, in); 
                }

                put_block(leadout, 0x0C);
	}
}


char *find_param(char *p, int *nr)
{
	char *q, *parameters[] = {"size", "hdpi", "mode", "bidi", NULL};
	int i;

	for (i=0; parameters[i]; i++) {
		if ( !strncmp(p, parameters[i], strlen(parameters[i])) ) {
			q = p + strlen(parameters[i]);
			if ( *q == '=') {
				*nr=i;
				return q+1;
			}
		}
	}

	*nr=-1;
	return NULL;
}


int main(int argc, char *argv[])
{
	int arg, i, parm;
	char *p;

	// process command line parameters
	for (arg=1; arg<argc; arg++) {
		p = find_param(argv[arg], &parm);
		switch(parm) {
			case 0:
				for (i=0; ghostscript[i].name; i++) {
					if ( !strcmp(ghostscript[i].name, p) ) {
						gs_width = ghostscript[i].width;
						gs_height = ghostscript[i].height;
						fprintf(stderr, "page size set to '%s'.\n", p);
						break;
					}
				}
				break;
			
			case 1:
				i=atoi(p);
				if (i==300 || i==600 || i==1200) {
					hdpi=i;
					fprintf(stderr, "horizontal resolution set to %d dpi.\n", i);
				}
				break;

			case 2:
				for (i=0; modenames[i]; i++) {
					if ( !strcmp(modenames[i], p) ) {
						mode=i;
						fprintf(stderr, "mode set to '%s'.\n", modenames[i]);
					}
				}
				break;

			case 3:
				bidirectional=atoi(p);
				fprintf(stderr, "bidirectional mode %s.\n", bidirectional ? "on" : "off");
				break;
		}
	}

	// start printing
	out = stdout;
	in = stdin;

	if (in && out) print();

	if (in) fclose(in);
	if (out) fclose(out);

	return 0;
}

// EOF

