/***************************************************************************
			      z42tool.c
                         -------------------
    version		 : 0.4.3
    date	         : September 14th 2004
    copyright            : (C) 2001,2002,2003 by Alwin Stolk
                           (C) 2003 Crispin Boylan (GTK2 patch)
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <gtk/gtk.h>
#include <signal.h>
#include <fcntl.h>
#include <unistd.h>
#include "z42tool.h"
#include "interface.h"
#include "support.h"

#define CARTRIDGE_UNKNOWN	0x10
#define LEFT_COLOUR_STD         0x50
#define LEFT_COLOUR_HIGH        0x60
#define RIGHT_BLACK_STD		0x30
#define RIGHT_BLACK_HIGH	0x40
#define RIGHT_PHOTO		0x70
#define NR_ALGN_COLOUR		6
#define NR_ALGN_PHOTO		4
#define PHOTOMODE		(rightcart==3)

#define	RDPR			0
#define WRPR			1

struct printer {
	char *name;
	int idlen;
	char *id;
	char *algn_std;
	char *algn_photo;
};

struct printer printers[] = {
	{ NULL, 0, NULL, "align_standard", "align_photo" },
	{ "Lexmark Z42",	226,	"Z42",	"lxaaalgn.out", "lxaaalgk.out" },
	{ "Lexmark Z43",	0,	"Z43",	"lxatalgn.out", "lxatalgk.out" }, // should work
	{ "Lexmark Z52",	0,	"Z52",	"lxacalgn.out", "lxacalgk.out" },
	{ "Lexmark Z53",        0,      "Z53",  "lxacalgn.out", "lxacalgk.out" },
	{ "Compaq IJ1200",	0,	"1200",	"lxaaalgn.out", "lxaaalgk.out" }, // relabeled Z42
	{ "Lexmark X73",	228,	"Z42",	"lxaralgn.out", "lxaralgk.out" },
	/*
	{ "Lexmark X83",	0,	"X83",	"lxasalgn.out", "lxasalgk.out" }, // not tested yet
	*/
	{ NULL, 0, NULL, NULL, NULL }
};
//enum printertypes { PRN_UNKNOWN, LEX_Z42, LEX_Z43, LEX_Z52, LEX_1200 };

typedef int bool;

unsigned char align_colour[NR_ALGN_COLOUR] = { 0,0,0,0,0,0 };
unsigned char align_photo[NR_ALGN_PHOTO] = { 0,0,0,0 };
int leftcart=0, rightcart=0;

// assume alignment files are (sym)links
// assume printer commands are compatible to Lexmark Z42
int prtype=0;

extern GtkWidget *window1, *cartridge_left, *cartridge_right;

void fetch_alignments(void);

/* ------------------------------------------------------------------- */

const char *get_device(void)
{
	GtkWidget *wid;

	wid = lookup_widget(GTK_WIDGET(window1), "printer_device");
	return (const char *) gtk_entry_get_text(GTK_ENTRY(wid));
}


void set_status(char *p)
{
	GtkWidget *wid;
	wid = lookup_widget(GTK_WIDGET(window1), "statusline");
	gtk_entry_set_text(GTK_ENTRY(wid), p);
}


void catcher(int signal)
{
	return;
}


ssize_t rwblk(unsigned char *data, size_t size, int timeout, bool wr)
{
	void *oldalarm;
	int fd;
	ssize_t len=0;

	oldalarm = signal(SIGALRM, catcher);
	if (oldalarm == SIG_ERR) return 0;

	fd = open( get_device(), wr ? O_WRONLY : O_RDONLY );
	if (fd >= 0) {
		alarm(timeout);
		if (wr) {
			len = write(fd, (void *)data, size);
		} else {
			len = read(fd, (void *)data, size);
		}
		alarm(0);
		close(fd);
	}

	signal(SIGALRM, oldalarm);
	return len; 
}


void print_alignment_page (GtkButton *button, gpointer user_data)
{
	FILE *in, *out;
	char *path = PACKAGE_DATA_DIR "/";
	char alignment_page[500];
	int c;
	GtkWidget *wid;
	bool ok = FALSE;

	read_printer_status(0,0);

	strcpy(alignment_page, path);
	strcat(alignment_page, PHOTOMODE ? printers[prtype].algn_photo : 
						printers[prtype].algn_std);

	in = fopen(alignment_page, "rb");
	if (in) {
		out = fopen(get_device(), "w+b");
		if (out) {
			while (1) {
				c = fgetc(in);
				if ( feof(in) ) break;
				fputc(c, out);
			}
			fclose(out);
			ok = TRUE;
		}
		fclose(in);
	} 

	if (ok) {
		set_status("Printed alignment page");
	} else {
		wid = lookup_widget(GTK_WIDGET(window1), "statusline");
		gtk_entry_set_text(GTK_ENTRY(wid), strerror(errno) );
		gtk_entry_append_text(GTK_ENTRY(wid), ": ");
		gtk_entry_append_text(GTK_ENTRY(wid), alignment_page);
	}
}


void align_cartridges (GtkButton *button, gpointer user_data)
{
	int x;
	unsigned char align[] = {
		0xA5,0x00,0x2F,0x40,0xE0,0x81,0x00,0x01,0x00,0x0A,0x03,0x00,0x01,0x10,0x03,0x01,
		0x01,0x06,0x03,0x02,0x01,0x0C,0x03,0x03,0x01,0x0C,0x03,0x04,0x01,0x0D,0x03,0x05,
		0x01,0x0B,0x03,0x20,0x01,0x0F,0x03,0x21,0x01,0x08,0x03,0x22,0x01,0x0F,0x03,0x23,
		0x01,0x0F
	};
	unsigned char algnc[] = { 0x0D, 0x11, 0x15, 0x19, 0x1D, 0x21 };
	unsigned char algnp[] = { 0x25, 0x29, 0x2D, 0x31 };

	fetch_alignments();

	for (x=0; x<NR_ALGN_COLOUR; x++) align[algnc[x]] = align_colour[x];
	for (x=0; x<NR_ALGN_PHOTO; x++) align[algnp[x]] = align_photo[x];

	if ( rwblk(align, sizeof(align), 5, WRPR) == sizeof(align) ) {
		set_status("Alignments sent to printer.");
	} else {
		set_status("Can't send alignments to printer.");
	}
}


void clean_print_nozzles (GtkButton *button, gpointer user_data)
{
	unsigned char headclean[] = {
	0x1b,0x2a,0x07,0x63,0x1b,0x2a,0x03,0x04,0xb0,0x1b,0x2a,0x6d,0x00,0x19,0x00,0x1b,
	0x2a,0x6d,0x00,0x19,0x01,0x1b,0x2a,0x6d,0x00,0x04,0x00,0x0a,0x00,0x01,0x1b,0x2a,
	0x6d,0x00,0x17,0x00,0x1b,0x2a,0x6d,0x00,0x17,0x01,0x1b,0x2a,0x6d,0x00,0x17,0x00,
	0x1b,0x2a,0x6d,0x00,0x17,0x01,0x1b,0x2a,0x6d,0x00,0x04,0x00,0x0a,0x00,0x01,0x1b,
	0x2a,0x6d,0x00,0x17,0x00,0x1b,0x2a,0x6d,0x00,0x17,0x01,0x1b,0x2a,0x6d,0x00,0x17,
	0x00,0x1b,0x2a,0x6d,0x00,0x17,0x01,0x1b,0x2a,0x6d,0x00,0x04,0x00,0x0a,0x00,0x01,
	0x1b,0x2a,0x6d,0x00,0x17,0x00,0x1b,0x2a,0x6d,0x00,0x17,0x01,0x1b,0x2a,0x6d,0x00,
	0x17,0x00,0x1b,0x2a,0x6d,0x00,0x17,0x01,0x1b,0x2a,0x6d,0x00,0x04,0x00,0x80,0x00,
	0x01,0x1b,0x2a,0x6d,0x00,0x19,0x00,0x1b,0x2a,0x6d,0x00,0x19,0x01,0x1b,0x2a,0x07,
	0x65,0x1b,0x2a,0x82,0x00,0x00,0x00,0x00,0xac
	};

	if ( rwblk(headclean, sizeof(headclean), 5, WRPR) == sizeof(headclean) ) {
		set_status("Nozzles cleaned.");
	} else {
		set_status("Can't clean nozzles.");
	}
}


void change_cartridges (GtkButton *button, gpointer user_data)
{
	unsigned char left_new=FALSE, right_new=FALSE;
	GtkWidget *wid;
	bool ok=FALSE;

	unsigned char change[] = {
		0xA5,0x00,0x17,0x40,0xE0,0x81,0x00,0x01,0x00,0x04,0x01,0x00,0x01,0x50,0x01,0x02,
		0x01,0x00,0x02,0x00,0x01,0x30,0x02,0x02,0x01,0x00
	};
	unsigned char no_full[] = {
		0xA5,0x00,0x07,0x40,0xE0,0x81,0x00,0x01,0x00,0x00
	};
	unsigned char left_full[] = {
		0xA5,0x00,0x1C,0x40,0xE0,0x81,0x00,0x01,0x00,0x03,0x04,0x01,0x04,0x00,0x00,0x00,
		0x00,0x04,0x02,0x04,0x00,0x00,0x00,0x00,0x04,0x03,0x04,0x00,0x00,0x00,0x00
	};
	unsigned char right_full[] = {
		0xA5,0x00,0x0E,0x40,0xE0,0x81,0x00,0x01,0x00,0x01,0x04,0x00,0x04,0x00,0x00,0x00,
		0x00
	};
	unsigned char both_full[] = {
		0xA5,0x00,0x23,0x40,0xE0,0x81,0x00,0x01,0x00,0x04,0x04,0x01,0x04,0x00,0x00,0x00,
		0x00,0x04,0x02,0x04,0x00,0x00,0x00,0x00,0x04,0x03,0x04,0x00,0x00,0x00,0x00,0x04,
		0x00,0x04,0x00,0x00,0x00,0x00
	};

	wid = lookup_widget(GTK_WIDGET(button), "full_left");
	if ( GTK_TOGGLE_BUTTON(wid)->active ) left_new = TRUE;
	wid = lookup_widget(GTK_WIDGET(button), "full_right");
	if ( GTK_TOGGLE_BUTTON(wid)->active ) right_new = TRUE;

	if (!leftcart || !rightcart) {
		set_status("Can't set cartridge type 'UNKNOWN'.");
		return;
	}

	switch(leftcart) {
		case 1:	change[0x0D] = LEFT_COLOUR_STD; break;
		case 2:	change[0x0D] = LEFT_COLOUR_HIGH; break;
	}
	switch(rightcart) {
		case 1:	change[0x15] = RIGHT_BLACK_STD; break;
		case 2:	change[0x15] = RIGHT_BLACK_HIGH; break;
		case 3: change[0x15] = RIGHT_PHOTO; break;
	}

	if ( rwblk(change, sizeof(change), 2, WRPR) == sizeof(change) ) {
		if (!left_new && !right_new) 
			ok = (rwblk(no_full, sizeof(no_full), 2, WRPR) == sizeof(no_full) );
		if ( left_new && !right_new) 
			ok = (rwblk(left_full, sizeof(left_full), 2, WRPR) == sizeof(left_full) );
		if (!left_new &&  right_new) 
			ok = (rwblk(right_full, sizeof(right_full), 2, WRPR) == sizeof(right_full) );
		if ( left_new &&  right_new) 
			ok = (rwblk(both_full, sizeof(both_full), 2, WRPR) == sizeof(both_full) );
	}

	set_status(ok ? "Cartridge types set." : "Can't set cartridge types.");
}


void update_parameters(void)
{
	GtkWidget *wid;
	int x;
	char alignment[] = "align_a";

	gtk_option_menu_set_history(GTK_OPTION_MENU(cartridge_left), leftcart);
	gtk_option_menu_set_history(GTK_OPTION_MENU(cartridge_right), rightcart);

	for (x=0; x<NR_ALGN_COLOUR; x++) {
		alignment[6] = 'a'+x;
		wid = lookup_widget(GTK_WIDGET(window1), alignment);
		if (PHOTOMODE) {
			gtk_spin_button_set_value(GTK_SPIN_BUTTON(wid), x<NR_ALGN_PHOTO ? align_photo[x] : 0);
		} else {
			gtk_spin_button_set_value(GTK_SPIN_BUTTON(wid), align_colour[x]);

		}
		gtk_widget_set_sensitive(wid, !PHOTOMODE || (x<NR_ALGN_PHOTO) );
	}
}


void fetch_alignments(void)
{
	GtkWidget *wid;
	char alignment[] = "align_a";
	int x, val;
	
	for (x=0; x<NR_ALGN_COLOUR; x++) {
		alignment[6] = 'a'+x;
		wid = lookup_widget(GTK_WIDGET(window1), alignment);
		val = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(wid));
		if (PHOTOMODE) {
			if (x<NR_ALGN_PHOTO) align_photo[x] = val;
		} else {
			align_colour[x] = val;
		}
	}
}


void set_left (GtkButton *button, gpointer user_data)
{
	leftcart = (int)user_data;
}


void set_right (GtkButton *button, gpointer user_data)
{
	fetch_alignments();
	rightcart = (int)user_data;
	update_parameters();
}


void read_printer_status (GtkButton *button, gpointer user_data)
{
	#define BUF_SZ		1000
	#define DUMP_STATUS	FALSE
	#define PRNID		0x90

	unsigned char request[] = { 0xA5, 0, 5, 0x50, 0xE0, 0x81, 0, 0xFF };
	unsigned char algnc[] = { 45, 49, 53, 57, 61, 70 };
	unsigned char algnp[] = { 74, 78, 82, 86 };
	unsigned char buf[BUF_SZ];
	int len, x;
	FILE *out;

	prtype = 0;
	rwblk(buf, BUF_SZ, 1, RDPR); // flush

	if ( rwblk(request, sizeof(request), 2, WRPR) == 8) {
		sleep(1);
		len = rwblk(buf, BUF_SZ, 5, RDPR);
		if (len==BUF_SZ) len--;
		buf[len] = 0;
		
		// debug
		if (len && DUMP_STATUS) {
			out = fopen("z42tool_dump","wb");
			if (out) {
				fwrite(buf, 1, len, out);
				fclose(out);
			}
		}

		// attempt identification
		if (len>PRNID) {
			for(x=1; printers[x].name; x++) {
				if (printers[x].idlen && (printers[x].idlen != len) ) continue;
				if ( strstr((char *)(buf+PRNID), printers[x].id) ) {
					prtype = x;
					break;
				}
			}
		}
	} 
	
	set_status(prtype ? printers[prtype].name : "Can't read printer status");

	// 'parse'
	if (prtype) {
		switch( buf[13] ) {
			case LEFT_COLOUR_STD:	leftcart = 1; break;
			case LEFT_COLOUR_HIGH:	leftcart = 2; break;
			default:		leftcart = 0; break;
		}
		switch( buf[29]) {
			case RIGHT_BLACK_STD:	rightcart = 1; break;
			case RIGHT_BLACK_HIGH:	rightcart = 2; break;
			case RIGHT_PHOTO:	rightcart = 3; break;
			default:		rightcart = 0; break;
		}

		for(x=0; x<NR_ALGN_COLOUR; x++) align_colour[x] = buf[algnc[x]];
		for(x=0; x<NR_ALGN_PHOTO; x++) align_photo[x] = buf[algnp[x]];

		update_parameters();
	}
}


void on_about(GtkButton *button, gpointer user_data)
{
	GtkWidget *about;

	about = create_dialog1();
	gtk_widget_show(about);
}


